#!/usr/bin/env python
"""
Usage:
    train.py [options] DATA_PATH MODEL_FILENAME

Options:
    --aml                      Run this in Azure ML
    --azure-info=<path>        Azure authentication information file (JSON). Used to load data from Azure storage.
    --max-num-epochs=<epochs>  The maximum number of epochs to run training for. [default: 1000]
    --minibatch-size=<size>    The minibatch size. [default: 20]
    --restore-path=<path>      The path to previous model file for starting from previous checkpoint.
    --sequential-run           Do not parallelize data loading. Makes debugging easier.
    --quiet                    Do not show progress bar.
    -h --help                  Show this screen.
    --debug                    Enable debug routines. [default: False]
"""
import json
import numpy as np
import torch
from docopt import docopt
from dpu_utils.utils import RichPath, run_and_debug
from pathlib import Path
from torch import nn

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModelTrainer
from ptgnn.baseneuralmodel.utils.amlutils import configure_logging, log_run
from ptgnn.implementations.ppi.dataloader import PPIDatasetLoader
from ptgnn.implementations.ppi.ppi import PPIMulticlassClassification
from ptgnn.neuralmodels.embeddings.linearmapembedding import FeatureRepresentationModel
from ptgnn.neuralmodels.gnn.graphneuralnetwork import GraphNeuralNetworkModel
from ptgnn.neuralmodels.gnn.messagepassing.mlpmessagepassing import MlpMessagePassingLayer
from ptgnn.neuralmodels.gnn.messagepassing.residuallayers import MeanResidualLayer


def create_ppi_gnn_model(hidden_state_size: int = 256):
    def create_mp_layers(num_edges: int):
        mlp_mp_constructor = lambda: MlpMessagePassingLayer(
            input_state_dimension=hidden_state_size,
            message_dimension=hidden_state_size,
            output_state_dimension=hidden_state_size,
            num_edge_types=num_edges,
            message_aggregation_function="sum",
            dropout_rate=0.2,
        )
        r1 = MeanResidualLayer(hidden_state_size)
        r2 = MeanResidualLayer(hidden_state_size)
        return [
            r1.pass_through_dummy_layer(),
            mlp_mp_constructor(),
            mlp_mp_constructor(),
            mlp_mp_constructor(),
            r1,
            r2.pass_through_dummy_layer(),
            mlp_mp_constructor(),
            mlp_mp_constructor(),
            r2,
        ]

    return PPIMulticlassClassification(
        gnn_model=GraphNeuralNetworkModel[np.ndarray, np.ndarray](
            node_representation_model=FeatureRepresentationModel(
                embedding_size=hidden_state_size,
                activation=nn.Tanh(),
            ),
            message_passing_layer_creator=create_mp_layers,
            max_nodes_per_graph=6000,
            max_graph_edges=300000,
            introduce_backwards_edges=True,
            add_self_edges=True,
            stop_extending_minibatch_after_num_nodes=3000,
        ),
    )


def run(arguments):
    if arguments["--aml"]:
        from azureml.core.run import Run

        aml_ctx = Run.get_context()
        assert torch.cuda.is_available(), "No CUDA available. Aborting training."
    else:
        aml_ctx = None

    log_path = configure_logging(aml_ctx)
    azure_info_path = arguments.get("--azure-info", None)

    data_path = RichPath.create(arguments["DATA_PATH"], azure_info_path)
    training_data = PPIDatasetLoader.load_data(data_path, "train")
    validation_data = PPIDatasetLoader.load_data(data_path, "valid")

    model_path = Path(arguments["MODEL_FILENAME"])
    assert model_path.name.endswith(".pkl.gz"), "MODEL_FILENAME must have a `.pkl.gz` suffix."

    initialize_metadata = True
    restore_path = arguments.get("--restore-path", None)

    if restore_path:
        initialize_metadata = False
        model, nn = AbstractNeuralModel.restore_model(Path(restore_path))
    else:
        model = create_ppi_gnn_model()
        nn = None

    def create_optimizer(parameters):
        return torch.optim.Adam(
            parameters,
            lr=1e-3,
        )

    trainer = ModelTrainer(
        model,
        model_path,
        max_num_epochs=int(arguments["--max-num-epochs"]),
        minibatch_size=int(arguments["--minibatch-size"]),
        optimizer_creator=create_optimizer,
        clip_gradient_norm=1,
        target_validation_metric="f1_score",
        target_validation_metric_higher_is_better=True,
    )
    if nn is not None:
        trainer.neural_module = nn

    trainer.register_train_epoch_end_hook(
        lambda model, nn, epoch, metrics: log_run(aml_ctx, "train", model, epoch, metrics)
    )
    trainer.register_validation_epoch_end_hook(
        lambda model, nn, epoch, metrics: log_run(aml_ctx, "valid", model, epoch, metrics)
    )

    trainer.train(
        training_data,
        validation_data,
        show_progress_bar=not arguments["--quiet"],
        initialize_metadata=initialize_metadata,
        parallelize=not arguments["--sequential-run"],
        patience=20,
    )

    test_data = PPIDatasetLoader.load_data(data_path, "test")
    metrics = model.report_metrics(
        test_data,
        trainer.neural_module,
        device=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"),
    )
    print(f"Test metrics: {json.dumps(metrics, indent=3)}")

    if aml_ctx is not None:
        aml_ctx.upload_file(name="model.pkl.gz", path_or_stream=str(model_path))
        aml_ctx.upload_file(name="full.log", path_or_stream=log_path)


if __name__ == "__main__":
    args = docopt(__doc__)
    run_and_debug(lambda: run(args), args.get("--debug", False))



import torch
from torch import nn
from torch_scatter import scatter_mean
from typing import Dict, List, Tuple

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import AbstractMessagePassingLayer


class GraphNorm(AbstractMessagePassingLayer):
    """
    A GraphNorm layer implemented as in

    GraphNorm: A Principled Approach to Accelerating Graph Neural Network Training (arXiv:2009.03294v1)
      Tianle Cai, Shengjie Luo, Keyulu Xu, Di He, Tie-yan Liu, Liwei Wang

    """

    def __init__(self, input_state_dimension: int, eps: float = 1e-10):
        super().__init__()
        self.__input_state_dim = input_state_dimension
        self.__eps = eps

        self.gamma = nn.Parameter(torch.ones(1, input_state_dimension))
        self.alpha = nn.Parameter(torch.ones(1, input_state_dimension))
        self.bias = nn.Parameter(torch.zeros(1, input_state_dimension))

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        per_graph_mean = scatter_mean(
            node_states, index=node_to_graph_idx, dim=0
        )  # [num_graphs, D]
        shifted = node_states - self.alpha * per_graph_mean[node_to_graph_idx]  # [num_nodes, D]
        sigma_2 = (
            scatter_mean(torch.pow(shifted, 2), index=node_to_graph_idx, dim=0) + self.__eps
        )  # [num_graphs, D]

        return (
            self.gamma * shifted / torch.sqrt(sigma_2[node_to_graph_idx]) + self.bias
        )  # [num_nodes, D]

    @property
    def input_state_dimension(self) -> int:
        return self.__input_state_dim

    @property
    def output_state_dimension(self) -> int:
        return self.__input_state_dim



import tempfile
import unittest
from pathlib import Path
from typing import List, Optional, Tuple

from ptgnn.baseneuralmodel import ModelTrainer
from ptgnn.tests.simplemodel.data import SampleDatapoint, SyntheticData
from ptgnn.tests.simplemodel.model import SimpleRegressionModel


class TestPytorchModel(unittest.TestCase):
    """
    Model of a Boolean classifier of positive/negative cross-product of input features and weights
    """

    def __get_data(
        self,
        num_points: int,
        num_features: int,
        random_seed: Optional[int] = None,
        train_test_split_pct=0.9,
    ) -> Tuple[List[SampleDatapoint], List[SampleDatapoint]]:
        """
        generates data as SyntheticData = SampleDataPoint[num_points] where
        SampleDataPoint<x,y> = <float[__num_features],bool>,  and y= sum(x*w) >0;
        weigths ~ N(0,1)*10, and SampleDataPoint.x ~ N(0,1)*5.
        Returns tuple of train and test data, split at @train_test_split_pct %
        """
        data = SyntheticData(num_features, random_seed=random_seed)
        all_data = list(data.generate(num_points))
        train_test_split = int(num_points * train_test_split_pct)
        training_data, validation_data = all_data[:train_test_split], all_data[train_test_split:]
        return training_data, validation_data

    def test_parallel(self):
        self.train_and_compare_model(True)

    def test_parallel_no_multiprocessing(self):
        self.train_and_compare_model(True, multiprocessing=False)

    def test_sequential(self):
        self.train_and_compare_model(False)

    def train_and_compare_model(self, parallelize: bool, multiprocessing: bool = True):
        num_points = 10000
        num_features = 100
        max_num_epochs = 50
        random_seed = 1234  # None to seed from clock
        training_data, validation_data = self.__get_data(
            num_points, num_features, random_seed=random_seed
        )

        with tempfile.TemporaryDirectory() as dir:
            model_file = Path(dir) / "testModel.pkl.gz"

            model = SimpleRegressionModel()
            trainer = ModelTrainer(model, model_file, max_num_epochs=max_num_epochs)
            trainer.train(
                training_data,
                validation_data,
                parallelize=parallelize,
                use_multiprocessing=multiprocessing,
            )
            model_acc_1 = model.compute_accuracy(
                trainer.neural_module,
                iter(validation_data),
                parallelize,
                use_multiprocessing=multiprocessing,
            )

            model, trained_network = SimpleRegressionModel.restore_model(model_file)
            trained_model_acc = model.compute_accuracy(
                trained_network,
                iter(validation_data),
                parallelize,
                use_multiprocessing=multiprocessing,
            )
            self.assertGreater(
                trained_model_acc, 0.95, f"Model achieves too low accuracy, {trained_model_acc:%}"
            )

            self.assertAlmostEqual(
                trained_model_acc,
                model_acc_1,
                places=3,
                msg=f"Accuracy before and after loading does not match: {trained_model_acc} vs {model_acc_1}",
            )


if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn
from typing import Callable, Dict, List, Optional, Tuple, Union

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import (
    AbstractMessageAggregation,
    AbstractMessagePassingLayer,
)
from ptgnn.neuralmodels.mlp import MLP


class MlpMessagePassingLayer(AbstractMessagePassingLayer):
    def __init__(
        self,
        input_state_dimension: int,
        output_state_dimension: int,
        message_dimension: int,
        num_edge_types: int,
        message_aggregation_function: Union[str, AbstractMessageAggregation],
        message_activation: Optional[nn.Module] = nn.GELU(),
        use_target_state_as_message_input: bool = True,
        mlp_hidden_layers: Union[List[int], int] = 0,
        use_layer_norm: bool = True,
        use_dense_layer: bool = True,
        dropout_rate: float = 0.0,
        dense_activation: Optional[nn.Module] = nn.Tanh(),
        features_dimension: int = 0,
    ):
        super().__init__()
        self.__input_state_dim = input_state_dimension
        self.__use_target_state_as_message_input = use_target_state_as_message_input
        self.__output_state_dim = output_state_dimension

        if use_target_state_as_message_input:
            message_input_size = 2 * input_state_dimension
        else:
            message_input_size = input_state_dimension
        self.__edge_message_transformation_layers = nn.ModuleList(
            [
                MLP(
                    input_dimension=message_input_size + features_dimension,
                    output_dimension=message_dimension,
                    hidden_layers=mlp_hidden_layers,
                )
                for _ in range(num_edge_types)
            ]
        )
        self.__aggregation_fn = message_aggregation_function
        if isinstance(message_aggregation_function, str):
            aggregated_state_size = message_dimension
        else:
            aggregated_state_size = self.__aggregation_fn.output_state_size(message_dimension)

        self.__message_activation = message_activation

        state_update_layers: List[nn.Module] = []
        if use_layer_norm:
            state_update_layers.append(nn.LayerNorm(aggregated_state_size))
        if use_dense_layer:
            state_update_layers.append(nn.Linear(aggregated_state_size, output_state_dimension))
            nn.init.xavier_uniform_(state_update_layers[-1].weight)
            if dense_activation is not None:
                state_update_layers.append(dense_activation)
        state_update_layers.append(nn.Dropout(p=dropout_rate))

        self.__state_update = nn.Sequential(*state_update_layers)

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        assert len(adjacency_lists) == len(self.__edge_message_transformation_layers)

        all_message_targets, all_messages = [], []
        for edge_type_idx, (adj_list, features, edge_transformation_layer) in enumerate(
            zip(adjacency_lists, edge_features, self.__edge_message_transformation_layers)
        ):
            edge_sources_idxs, edge_target_idxs = adj_list
            all_message_targets.append(edge_target_idxs)

            edge_source_states = nn.functional.embedding(edge_sources_idxs, node_states)

            if self.__use_target_state_as_message_input:
                edge_target_states = nn.functional.embedding(edge_target_idxs, node_states)
                message_input = torch.cat([edge_source_states, edge_target_states], dim=-1)
            else:
                message_input = edge_source_states

            all_messages.append(
                edge_transformation_layer(torch.cat([message_input, features], dim=-1))
            )

        if isinstance(self.__aggregation_fn, AbstractMessageAggregation):
            aggregated_messages = self.__aggregation_fn(
                messages=torch.cat(all_messages, dim=0),
                message_targets=torch.cat(all_message_targets, dim=0),
                num_nodes=node_states.shape[0],
            )
        else:
            aggregated_messages = self._aggregate_messages(
                messages=torch.cat(all_messages, dim=0),
                message_targets=torch.cat(all_message_targets, dim=0),
                num_nodes=node_states.shape[0],
                aggregation_fn=self.__aggregation_fn,
            )

        if self.__message_activation is not None:
            aggregated_messages = self.__message_activation(aggregated_messages)

        return self.__state_update(aggregated_messages)  # num_nodes x H

    @property
    def input_state_dimension(self) -> int:
        return self.__input_state_dim

    @property
    def output_state_dimension(self) -> int:
        return self.__output_state_dim



import torch
from torch import nn
from typing import Dict, List, Tuple

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import AbstractMessagePassingLayer


class GatedMessagePassingLayer(AbstractMessagePassingLayer):
    def __init__(
        self,
        state_dimension: int,
        message_dimension: int,
        num_edge_types: int,
        message_aggregation_function: str,
        dropout_rate: float = 0.0,
        edge_feature_dimension: int = 0,
    ):
        super().__init__()

        self.__edge_message_transformation_layers = nn.ModuleList(
            [
                nn.Linear(state_dimension + edge_feature_dimension, message_dimension, bias=False)
                for _ in range(num_edge_types)
            ]
        )
        for msg_layer in self.__edge_message_transformation_layers:
            nn.init.xavier_normal_(msg_layer.weight, gain=(1 / num_edge_types) ** 0.5)
        self.__state_update = nn.GRUCell(input_size=message_dimension, hidden_size=state_dimension)
        nn.init.orthogonal_(self.__state_update.weight_hh)
        nn.init.xavier_uniform_(self.__state_update.weight_ih)
        nn.init.normal_(self.__state_update.bias_hh, std=1e-5)
        nn.init.normal_(self.__state_update.bias_ih, std=1e-5)
        self.__state_dimension = state_dimension
        self.__aggregation_fn = message_aggregation_function
        self.__dropout = nn.Dropout(p=dropout_rate)

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        message_targets = torch.cat([adj_list[1] for adj_list in adjacency_lists])  # num_messages
        assert len(adjacency_lists) == len(self.__edge_message_transformation_layers)

        all_messages = []
        for edge_type_idx, (adj_list, features, edge_transformation_layer) in enumerate(
            zip(adjacency_lists, edge_features, self.__edge_message_transformation_layers)
        ):
            edge_sources_idxs = adj_list[0]
            edge_source_states = nn.functional.embedding(
                edge_sources_idxs, node_states
            )  # [num_edges_of_type_edge_type_idx, H]
            all_messages.append(
                edge_transformation_layer(
                    self.__dropout(torch.cat([edge_source_states, features], -1))
                )
            )  # [num_edges_of_type_edge_type_idx, D]

        aggregated_messages = self._aggregate_messages(
            messages=torch.cat(all_messages, dim=0),
            message_targets=message_targets,
            num_nodes=node_states.shape[0],
            aggregation_fn=self.__aggregation_fn,
        )  # [num_nodes, D]
        return self.__state_update(aggregated_messages, node_states)  # [num_nodes, H]

    @property
    def input_state_dimension(self) -> int:
        return self.__state_dimension

    @property
    def output_state_dimension(self) -> int:
        return self.__state_dimension



import logging
import os
import sys
from typing import Any, Dict


def configure_logging(aml_ctx) -> str:
    os.makedirs("logs", exist_ok=True)

    base_logger = logging.getLogger()
    base_logger.setLevel(logging.INFO)

    log_path = os.path.join("logs", "full.log")
    formatter = logging.Formatter(
        "%(asctime)s [%(name)-35.35s @ %(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s"
    )
    file_handler = logging.FileHandler(log_path)
    file_handler.setFormatter(formatter)
    base_logger.addHandler(file_handler)

    # if aml_ctx is None:
    #     # AML seems to be adding its own logger behind the scenes, no need to add another one.
    stream_handler = logging.StreamHandler(sys.stdout)
    stream_handler.setFormatter(formatter)
    base_logger.addHandler(stream_handler)

    return log_path


def log_run(aml_ctx, fold_name: str, model, epoch_idx: int, metrics: Dict[str, Any]) -> None:
    if aml_ctx is None:
        return
    for metric_name, metric_value in metrics.items():
        aml_ctx.log(f"{fold_name}-{metric_name}", metric_value)



import torch
from abc import abstractmethod
from torch import nn
from typing import Dict, List, Tuple

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import AbstractMessagePassingLayer
from ptgnn.neuralmodels.reduceops.varsizedsummary import (
    AbstractVarSizedElementReduce,
    ElementsToSummaryRepresentationInput,
)


class AbstractGlobalGraphExchange(AbstractMessagePassingLayer):
    def __init__(
        self,
        global_graph_representation_module: AbstractVarSizedElementReduce,
        dropout_rate: float = 0.0,
    ):
        super().__init__()
        self.__global_graph_representation_module = global_graph_representation_module
        self.__dropout = nn.Dropout(p=dropout_rate)

    @abstractmethod
    def _update_node_states(
        self, node_states: torch.Tensor, global_info_per_node: torch.Tensor
    ) -> torch.Tensor:
        raise NotImplementedError()

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        e = ElementsToSummaryRepresentationInput(
            element_embeddings=node_states,
            element_to_sample_map=node_to_graph_idx,
            num_samples=node_to_graph_idx.max() + 1,
        )
        graph_representations = self.__global_graph_representation_module(e)
        graph_representations = self.__dropout(graph_representations)
        return self._update_node_states(node_states, graph_representations[node_to_graph_idx])


class GruGlobalStateUpdate(AbstractGlobalGraphExchange):
    def __init__(
        self,
        global_graph_representation_module: AbstractVarSizedElementReduce,
        input_state_size: int,
        summarized_state_size: int,
        dropout_rate: float = 0.0,
    ):
        super().__init__(global_graph_representation_module, dropout_rate)
        self.__input_dim = input_state_size
        self.__summarized_state_size = summarized_state_size
        self.__gru_cell = nn.GRUCell(input_size=summarized_state_size, hidden_size=input_state_size)

    def _update_node_states(
        self, node_states: torch.Tensor, global_info_per_node: torch.Tensor
    ) -> torch.Tensor:
        return self.__gru_cell(global_info_per_node, node_states)

    @property
    def input_state_dimension(self) -> int:
        return self.__input_dim

    @property
    def output_state_dimension(self) -> int:
        return self.__input_dim



import random
from typing import Callable, Iterable, Iterator, List, Optional, TypeVar

T = TypeVar("T")


class LazyDataIterable(Iterable[T]):
    def __init__(self, base_iterable_func: Callable[[], Iterator[T]]):
        self.__base_iterable_func = base_iterable_func

    def __iter__(self) -> Iterator[T]:
        return self.__base_iterable_func()


class MemorizedDataIterable(Iterable[T]):
    def __init__(self, base_iterable_func: Callable[[], Iterator[T]], shuffle: bool = False):
        self.__base_iterable_func = base_iterable_func
        self.__elements: List[T] = []
        self.__use_cache: bool = False
        self.__shuffle = shuffle

    def __yield_and_store(self, base: Iterator[T]):
        for element in base:
            self.__elements.append(element)
            yield element
        self.__use_cache = True
        del self.__base_iterable_func

    def __iter__(self) -> Iterator[T]:
        if self.__use_cache:
            if self.__shuffle:
                random.shuffle(self.__elements)
            return iter(self.__elements)
        else:
            return self.__yield_and_store(self.__base_iterable_func())

    def __call__(self) -> Iterator[T]:
        return iter(self)


def enforce_not_None(e: Optional[T]) -> T:
    """Enforce non-nullness of input. Used for typechecking and runtime safety."""
    if e is None:
        raise Exception("Input is None.")
    return e



import torch
from torch import nn
from torch_scatter import scatter_sum
from typing import Dict, List, Tuple

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import AbstractMessagePassingLayer


class MultiHeadSelfAttentionMessagePassing(AbstractMessagePassingLayer):
    """
    A transformer layer among all nodes in a graph.
    """

    def __init__(
        self,
        input_state_dimension: int,
        key_query_dimension: int,
        value_dimension: int,
        output_dimension: int,
        intermediate_dimension: int,
        num_heads: int,
        dropout_rate: float = 0.0,
        target_reference: str = "all",
        max_num_nodes: int = 250,
    ):
        super().__init__()

        self.__num_heads = num_heads
        self.__key_query_dim = key_query_dimension
        self.__value_dim = value_dimension

        self.__selfatt_head_transforms = nn.Linear(
            in_features=input_state_dimension,
            out_features=num_heads * (2 * key_query_dimension + value_dimension),
            bias=False,
        )

        self.__summarization_layer = nn.Linear(
            in_features=num_heads * value_dimension,
            out_features=output_dimension,
            bias=False,
        )

        self.__intermediate_layer = nn.Linear(
            in_features=output_dimension, out_features=intermediate_dimension
        )

        self.__output_layer = nn.Linear(
            in_features=intermediate_dimension, out_features=output_dimension
        )

        self.__layer_norm1 = nn.LayerNorm(output_dimension)
        self.__layer_norm2 = nn.LayerNorm(output_dimension)
        self.__dropout_layer = nn.Dropout(p=dropout_rate)

        self.__target_reference = target_reference
        self.__max_num_nodes = max_num_nodes

    def __iter_idxs_per_graph(self, node_to_graph_idx):
        with torch.no_grad():
            num_nodes_per_graph = scatter_sum(
                src=torch.ones_like(node_to_graph_idx, dtype=torch.int64), index=node_to_graph_idx
            )  # [num_nodes]

            node_offset = 0
            for num_nodes in num_nodes_per_graph:
                for start_idx in range(0, num_nodes, self.__max_num_nodes):
                    end_idx = min(start_idx + self.__max_num_nodes, num_nodes)
                    yield torch.arange(
                        start=start_idx,
                        end=end_idx,
                        dtype=torch.int64,
                        device=num_nodes_per_graph.device,
                    ) + node_offset
                node_offset += num_nodes

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],  # Not used
    ) -> torch.Tensor:
        if self.__target_reference == "all":
            relevant_node_states = node_states
        else:
            relevant_node_states = node_states[reference_node_ids[self.__target_reference]]
            node_to_graph_idx = reference_node_graph_idx[self.__target_reference]

        keys_queries_values = self.__selfatt_head_transforms(relevant_node_states).reshape(
            relevant_node_states.shape[0], self.__num_heads, -1
        )

        keys = keys_queries_values[:, :, : self.__key_query_dim]  # [num_nodes, num_heads, key_dim]
        queries = keys_queries_values[
            :, :, self.__key_query_dim : 2 * self.__key_query_dim
        ]  # [num_nodes, num_heads, key_dim]
        values = keys_queries_values[
            :, :, 2 * self.__key_query_dim :
        ]  # [num_nodes, num_heads, value_dim]

        all_out_values = []
        for graph_nodes in self.__iter_idxs_per_graph(node_to_graph_idx):
            # Loop is necessary due to limited memory and ease of coding.
            graph_keys = keys[graph_nodes]
            graph_queries = queries[graph_nodes]
            scores = torch.einsum("khd,vhd->khv", graph_keys, graph_queries) / (
                self.__key_query_dim ** 0.5
            )
            attention_probs = nn.functional.softmax(scores, dim=-1)
            attention_probs = self.__dropout_layer(attention_probs)
            out_values = torch.einsum("khv,vhd->khd", attention_probs, values[graph_nodes])
            all_out_values.append(out_values)

        values = torch.cat(all_out_values, dim=0)
        output = self.__summarization_layer(values.reshape(values.shape[0], -1))
        attention_output = self.__layer_norm1(self.__dropout_layer(output) + node_states)

        intermediate = nn.functional.relu(self.__intermediate_layer(attention_output))
        output = self.__dropout_layer(self.__output_layer(intermediate))
        output_node_states = self.__layer_norm2(output + attention_output)
        if self.__target_reference == "all":
            return output_node_states
        else:
            node_states[reference_node_ids[self.__target_reference]] = output_node_states
            return node_states

    @property
    def input_state_dimension(self) -> int:
        return self.__selfatt_head_transforms.in_features

    @property
    def output_state_dimension(self) -> int:
        return self.__output_layer.out_features



from typing_extensions import Final

import numpy as np
import torch
from torch import nn
from typing import Any, Dict, Optional, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel
from ptgnn.baseneuralmodel.utils.data import enforce_not_None
from ptgnn.neuralmodels.gnn.structs import AbstractNodeEmbedder


class LinearFeatureEmbedder(nn.Module):
    def __init__(
        self,
        input_element_size: int,
        output_embedding_size: int,
        activation: Optional[nn.Module] = None,
    ):
        super().__init__()
        self.__linear_map = nn.Linear(input_element_size, output_embedding_size, bias=False)
        torch.nn.init.xavier_uniform(self.__linear_map.weight)
        self.__activation = activation

    def forward(self, features):
        mapped_features = self.__linear_map(features)
        if self.__activation is not None:
            mapped_features = self.__activation(mapped_features)
        return mapped_features


class FeatureRepresentationModel(
    AbstractNeuralModel[np.ndarray, np.ndarray, LinearFeatureEmbedder],
    AbstractNodeEmbedder,
):
    """
    A model that maps a feature array to a D-sized representation (embedding) using a single linear layer.
    """

    def __init__(
        self,
        *,
        embedding_size: int = 64,
        activation: Optional[nn.Module] = None,
    ):
        super().__init__()
        self.embedding_size: Final = embedding_size
        self.__activation: Final = activation

    def representation_size(self) -> int:
        return self.embedding_size

    def initialize_metadata(self) -> None:
        self.__num_input_features = None

    def update_metadata_from(self, datapoint: np.ndarray) -> None:
        if self.__num_input_features is None:
            self.__num_input_features = datapoint.shape[0]
        else:
            assert (
                self.__num_input_features == datapoint.shape[0]
            ), "All samples should have the same number of features."

    def build_neural_module(self) -> LinearFeatureEmbedder:
        return LinearFeatureEmbedder(
            input_element_size=enforce_not_None(self.__num_input_features),
            output_embedding_size=self.embedding_size,
            activation=self.__activation,
        )

    def tensorize(self, datapoint: np.ndarray) -> np.ndarray:
        return datapoint

    def initialize_minibatch(self) -> Dict[str, Any]:
        return {"features": []}

    def extend_minibatch_with(
        self, tensorized_datapoint: np.ndarray, partial_minibatch: Dict[str, Any]
    ) -> bool:
        partial_minibatch["features"].append(tensorized_datapoint)
        return True

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        return {
            "features": torch.tensor(
                accumulated_minibatch_data["features"], dtype=torch.float32, device=device
            )
        }



import torch
from torch import nn
from typing import Dict, List, Tuple

from ptgnn.neuralmodels.gnn.messagepassing.abstractmessagepassing import AbstractMessagePassingLayer


class _ResidualOriginLayer(AbstractMessagePassingLayer):
    def __init__(self, input_dim: int, target_layer):
        super().__init__()
        self.__target_layer = target_layer
        self.__input_dim = input_dim

    @property
    def input_state_dimension(self) -> int:
        return self.__input_dim

    @property
    def output_state_dimension(self) -> int:
        return self.__input_dim

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        self.__target_layer._original_input = node_states
        return node_states


class MeanResidualLayer(AbstractMessagePassingLayer):
    def __init__(self, input_dim: int):
        super().__init__()
        self._original_input = None
        self.__input_dim = input_dim

    def pass_through_dummy_layer(self) -> _ResidualOriginLayer:
        return _ResidualOriginLayer(self.__input_dim, target_layer=self)

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        assert self._original_input is not None, "Initial Pass Through Layer was not used."
        out = torch.stack((self._original_input, node_states), dim=-1).mean(dim=-1)
        self._original_input = None  # Reset
        return out

    @property
    def input_state_dimension(self) -> int:
        return self.__input_dim

    @property
    def output_state_dimension(self) -> int:
        return self.__input_dim


class ConcatResidualLayer(AbstractMessagePassingLayer):
    def __init__(self, input_dim: int):
        super().__init__()
        self._original_input = None
        self.__input_dim = input_dim

    def pass_through_dummy_layer(self) -> _ResidualOriginLayer:
        return _ResidualOriginLayer(self.__input_dim, target_layer=self)

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        assert self._original_input is not None, "Initial Pass Through Layer was not used."
        out = torch.cat((self._original_input, node_states), dim=-1)
        self._original_input = None  # Reset
        return out

    @property
    def input_state_dimension(self) -> int:
        return self.__input_dim

    @property
    def output_state_dimension(self) -> int:
        return 2 * self.__input_dim


class LinearResidualLayer(AbstractMessagePassingLayer):
    def __init__(
        self,
        state_dimension1: int,
        state_dimension2: int,
        target_state_size: int,
        dropout_rate: float = 0.0,
    ):
        super().__init__()
        self.__input_dim1 = state_dimension1
        self.__input_dim2 = state_dimension2

        self.__linear_combination = nn.Linear(
            in_features=state_dimension1 + state_dimension2,
            out_features=target_state_size,
            bias=False,
        )
        self._original_input = None
        self.__dropout = nn.Dropout(p=dropout_rate)

    def pass_through_dummy_layer(self) -> _ResidualOriginLayer:
        return _ResidualOriginLayer(self.__input_dim1, target_layer=self)

    def forward(
        self,
        node_states: torch.Tensor,
        adjacency_lists: List[Tuple[torch.Tensor, torch.Tensor]],
        node_to_graph_idx: torch.Tensor,
        reference_node_ids: Dict[str, torch.Tensor],
        reference_node_graph_idx: Dict[str, torch.Tensor],
        edge_features: List[torch.Tensor],
    ) -> torch.Tensor:
        assert self._original_input is not None, "Initial Pass Through Layer was not used."
        out = self.__linear_combination(torch.cat((self._original_input, node_states), axis=-1))
        self._original_input = None  # Reset
        return self.__dropout(out)

    @property
    def input_state_dimension(self) -> int:
        return self.__input_dim2

    @property
    def output_state_dimension(self) -> int:
        return self.__linear_combination.out_features



import os
from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    long_description = f.read()

setup(
    name="ptgnn",
    packages=find_packages(),
    license="MIT",
    package_dir={"ptgnn": "ptgnn"},
    test_suite="ptgnn.tests",
    python_requires=">=3.6.1",
    description="Graph Neural Network library for PyTorch",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Deep Procedural Intelligence",
    package_data={"ptgnn": ["py.typed"]},
    install_requires=[
        "dpu-utils>=0.2.17",
        "jellyfish",
        "numpy",
        "torch-scatter>=2.0.4",
        "torch>=1.4.0",
        "tqdm",
        "typing-extensions",
    ],
    extras_require={"dev": ["black", "isort", "pre-commit"], "aml": ["azureml"]},
    setup_requires=["setuptools_scm"],
    url="https://github.com/microsoft/ptgnn/",
    project_urls={
        "Bug Tracker": "https://github.com/microsoft/ptgnn/issues",
        "Documentation": "https://github.com/microsoft/ptgnn/tree/master/docs",
        "Source Code": "https://github.com/microsoft/ptgnn",
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Libraries",
        "Typing :: Typed",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    zip_safe=False,
)



#!/usr/bin/env python
"""
Usage:
    train.py [options] TRAIN_DATA_PATH VALID_DATA_PATH MODEL_FILENAME

Options:
    --aml                      Run this in Azure ML
    --amp                      Enable automatic mixed precision.
    --azure-info=<path>        Azure authentication information file (JSON). Used to load data from Azure storage.
    --max-num-epochs=<epochs>  The maximum number of epochs to run training for. [default: 100]
    --minibatch-size=<size>    The minibatch size. [default: 300]
    --restore-path=<path>      The path to previous model file for starting from previous checkpoint.
    --sequential-run           Do not parallelize data loading. Makes debugging easier.
    --quiet                    Do not show progress bar.
    -h --help                  Show this screen.
    --debug                    Enable debug routines. [default: False]
"""
from docopt import docopt
from dpu_utils.utils import RichPath, run_and_debug
from pathlib import Path

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModelTrainer
from ptgnn.baseneuralmodel.utils.amlutils import configure_logging, log_run
from ptgnn.baseneuralmodel.utils.data import LazyDataIterable
from ptgnn.implementations.graph2seq.graph2seq import Graph2Seq
from ptgnn.neuralmodels.embeddings.strelementrepresentationmodel import (
    StrElementRepresentationModel,
)
from ptgnn.neuralmodels.gnn.graphneuralnetwork import GraphNeuralNetworkModel
from ptgnn.neuralmodels.gnn.messagepassing.gatedmessagepassing import GatedMessagePassingLayer
from ptgnn.neuralmodels.gnn.messagepassing.residuallayers import MeanResidualLayer
from ptgnn.neuralmodels.sequence.grucopydecoder import GruCopyingDecoderModel


def run(arguments):
    if arguments["--aml"]:
        import torch
        from azureml.core.run import Run

        aml_ctx = Run.get_context()
        assert torch.cuda.is_available(), "No CUDA available. Aborting training."
    else:
        aml_ctx = None

    log_path = configure_logging(aml_ctx)
    azure_info_path = arguments.get("--azure-info", None)

    training_data_path = RichPath.create(arguments["TRAIN_DATA_PATH"], azure_info_path)
    training_data = LazyDataIterable(lambda: training_data_path.read_as_jsonl())

    validation_data_path = RichPath.create(arguments["VALID_DATA_PATH"], azure_info_path)
    validation_data = LazyDataIterable(lambda: validation_data_path.read_as_jsonl())

    model_path = Path(arguments["MODEL_FILENAME"])
    assert model_path.name.endswith(".pkl.gz"), "MODEL_FILENAME must have a `.pkl.gz` suffix."

    initialize_metadata = True
    restore_path = arguments.get("--restore-path", None)
    if restore_path:
        initialize_metadata = False
        model, nn = AbstractNeuralModel.restore_model(Path(restore_path))
    else:
        embedding_size = 128
        dropout_rate = 0.1
        nn = None

        def create_mp_layers(num_edges: int):
            ggnn_mp = GatedMessagePassingLayer(
                state_dimension=embedding_size,
                message_dimension=embedding_size,
                num_edge_types=num_edges,
                message_aggregation_function="sum",
                dropout_rate=dropout_rate,
            )
            r1 = MeanResidualLayer(embedding_size)
            return [
                r1.pass_through_dummy_layer(),
                ggnn_mp,
                ggnn_mp,
                ggnn_mp,
                ggnn_mp,
                ggnn_mp,
                ggnn_mp,
                ggnn_mp,
                r1,
                GatedMessagePassingLayer(
                    state_dimension=embedding_size,
                    message_dimension=embedding_size,
                    num_edge_types=num_edges,
                    message_aggregation_function="sum",
                    dropout_rate=dropout_rate,
                ),
            ]

        model = Graph2Seq(
            gnn_model=GraphNeuralNetworkModel(
                node_representation_model=StrElementRepresentationModel(
                    token_splitting="token",
                    embedding_size=embedding_size,
                ),
                message_passing_layer_creator=create_mp_layers,
            ),
            decoder=GruCopyingDecoderModel(
                hidden_size=128, embedding_size=256, memories_hidden_dim=embedding_size
            ),
        )

    trainer = ModelTrainer(
        model,
        model_path,
        max_num_epochs=int(arguments["--max-num-epochs"]),
        minibatch_size=int(arguments["--minibatch-size"]),
        enable_amp=arguments["--amp"],
    )
    if nn is not None:
        trainer.neural_module = nn

    trainer.register_train_epoch_end_hook(
        lambda model, nn, epoch, metrics: log_run(aml_ctx, "train", model, epoch, metrics)
    )
    trainer.register_validation_epoch_end_hook(
        lambda model, nn, epoch, metrics: log_run(aml_ctx, "valid", model, epoch, metrics)
    )

    trainer.train(
        training_data,
        validation_data,
        show_progress_bar=not arguments["--quiet"],
        initialize_metadata=initialize_metadata,
        parallelize=not arguments["--sequential-run"],
    )

    if aml_ctx is not None:
        aml_ctx.upload_file(name="model.pkl.gz", path_or_stream=str(model_path))
        aml_ctx.upload_file(name="full.log", path_or_stream=log_path)


if __name__ == "__main__":
    args = docopt(__doc__)
    run_and_debug(lambda: run(args), args.get("--debug", False))



from typing_extensions import Final

import logging
import torch
from torch import nn
from typing import List, Optional, Union


class MLP(nn.Module):
    LOGGER: Final = logging.getLogger(__name__)

    def __init__(
        self,
        input_dimension: int,
        output_dimension: int,
        hidden_layers: Union[List[int], int] = 1,
        use_biases: bool = False,
        activation: Optional[nn.Module] = nn.ReLU(),
        dropout_rate: float = 0.0,
    ):
        """
        :param input_dimension: Dimensionality of input.
        :param hidden_layers: Either an integer determining number of hidden layers, which
                will have output_dimension units each; or list of integers whose lengths
                determines the number of hidden layers and whose contents the number of units in
                each layer.
        :param output_dimension: Dimensionality of output.
        :param use_biases: Flag indicating use of bias in fully connected layers.
        :param activation: Activation function applied between hidden layers (NB: the output of the
                MLP is always the direct result of a linear transformation)
        :param dropout_rate: Dropout applied to inputs of each MLP layer.
        """
        super().__init__()
        if isinstance(hidden_layers, int):
            if output_dimension == 1:
                self.LOGGER.warning(
                    f"W: MLP was created with {hidden_layers} layers of size 1, which is most likely wrong."
                    f" Switching to {hidden_layers} layers of size 32; to get hidden layers of size 1,"
                    f" use hidden_layers=[1,...,1] explicitly.",
                )
                hidden_layer_sizes = [32] * hidden_layers
            else:
                hidden_layer_sizes = [output_dimension] * hidden_layers
        else:
            hidden_layer_sizes = hidden_layers

        if len(hidden_layer_sizes) > 1:
            assert activation is not None, "Multiple linear layers without an activation"

        layers: List[nn.Module] = []
        cur_in_dim = input_dimension
        for hidden_layer_size in hidden_layer_sizes:
            layers.append(nn.Dropout(p=dropout_rate))
            layers.append(
                nn.Linear(
                    in_features=cur_in_dim,
                    out_features=hidden_layer_size,
                    bias=use_biases,
                )
            )
            nn.init.xavier_uniform_(layers[-1].weight)
            if activation is not None:
                layers.append(activation)
            cur_in_dim = hidden_layer_size

        # Sdd the final layer, but no final activation:
        layers.append(nn.Dropout(p=dropout_rate))
        layers.append(
            nn.Linear(
                in_features=cur_in_dim,
                out_features=output_dimension,
                bias=use_biases,
            )
        )
        nn.init.xavier_uniform_(layers[-1].weight)

        self.__mlp_modules = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.__mlp_modules(x)



from typing_extensions import Final, TypedDict

import torch
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

from ptgnn.baseneuralmodel import AbstractNeuralModel, ModuleWithMetrics
from ptgnn.baseneuralmodel.utils.data import enforce_not_None
from ptgnn.neuralmodels.gnn.graphneuralnetwork import GraphNeuralNetwork, GraphNeuralNetworkModel
from ptgnn.neuralmodels.gnn.structs import GnnOutput, GraphData, TensorizedGraphData
from ptgnn.neuralmodels.reduceops import (
    AbstractVarSizedElementReduce,
    ElementsToSummaryRepresentationInput,
    MultiheadSelfAttentionVarSizedElementReduce,
    SimpleVarSizedElementReduce,
)
from ptgnn.neuralmodels.sequence.grucopydecoder import (
    DecoderData,
    GruCopyingDecoder,
    GruCopyingDecoderModel,
    TokenizedOutput,
)


class CodeGraph2Seq(TypedDict):
    backbone_sequence: List[int]
    node_labels: List[str]
    edges: Dict[str, List[Tuple[int, int]]]
    method_name: List[str]


class TensorizedGraph2Seq(NamedTuple):
    encoder_data: TensorizedGraphData
    decoder_data: TokenizedOutput


class Graph2SeqModule(ModuleWithMetrics):
    def __init__(
        self,
        gnn: GraphNeuralNetwork,
        decoder: GruCopyingDecoder,
        node_to_graph_representation: AbstractVarSizedElementReduce,
    ):
        super().__init__()
        self._gnn = gnn
        self._decoder = decoder
        self.__node_to_graph_representation = node_to_graph_representation

    def _reset_module_metrics(self) -> None:
        self.__loss_sum = 0.0
        self.__num_mbs = 0

    def _module_metrics(self) -> Dict[str, Any]:
        return {"loss": self.__loss_sum / self.__num_mbs}

    def _get_initial_decoder_states(self, gnn_output: GnnOutput):
        return self.__node_to_graph_representation(
            ElementsToSummaryRepresentationInput(
                element_embeddings=torch.cat(
                    (gnn_output.input_node_representations, gnn_output.output_node_representations),
                    dim=-1,
                ),
                element_to_sample_map=gnn_output.node_to_graph_idx,
                num_samples=gnn_output.num_graphs,
            )
        )

    def forward(self, *, encoder_mb_data: Dict[str, Any], decoder_mb_data: Dict[str, Any]):
        gnn_output: GnnOutput = self._gnn(**encoder_mb_data)

        loss = self._decoder(
            input_memories=gnn_output.output_node_representations[
                gnn_output.node_idx_references["backbone_nodes"]
            ],
            input_memories_origin_idx=gnn_output.node_graph_idx_reference["backbone_nodes"],
            initial_states=self._get_initial_decoder_states(gnn_output),
            **decoder_mb_data
        )
        with torch.no_grad():
            self.__loss_sum += float(loss.cpu())
            self.__num_mbs += 1
        return loss


class Graph2Seq(AbstractNeuralModel[CodeGraph2Seq, TensorizedGraph2Seq, Graph2SeqModule]):
    def __init__(
        self,
        gnn_model: GraphNeuralNetworkModel,
        decoder: GruCopyingDecoderModel,
        num_summarization_heads: int = 8,
    ):
        super().__init__()
        self.__gnn_model = gnn_model
        self.__decoder_model = decoder
        self.num_summarization_heads: Final = num_summarization_heads

    def update_metadata_from(self, datapoint: CodeGraph2Seq) -> None:
        graph_nodes = [l.lower() for l in datapoint["node_labels"]]
        self.__gnn_model.update_metadata_from(
            GraphData(
                node_information=graph_nodes,
                edges=datapoint["edges"],
                reference_nodes={"backbone_nodes": datapoint["backbone_sequence"]},
            )
        )

        self.__decoder_model.update_metadata_from(
            DecoderData(
                input_elements=[graph_nodes[k] for k in datapoint["backbone_sequence"]],
                target_data=datapoint["method_name"],
            ),
        )

    def build_neural_module(self) -> Graph2SeqModule:
        gnn = self.__gnn_model.build_neural_module()
        decoder = self.__decoder_model.build_neural_module()
        node_to_graph_representation = MultiheadSelfAttentionVarSizedElementReduce(
            input_representation_size=gnn.input_node_state_dim + gnn.output_node_state_dim,
            hidden_size=gnn.input_node_state_dim + gnn.output_node_state_dim,
            output_representation_size=gnn.output_node_state_dim,
            num_heads=self.num_summarization_heads,
            query_representation_summarizer=SimpleVarSizedElementReduce("max"),
        )
        return Graph2SeqModule(gnn, decoder, node_to_graph_representation)

    def tensorize(self, datapoint: CodeGraph2Seq) -> Optional[TensorizedGraph2Seq]:
        graph_nodes = [l.lower() for l in datapoint["node_labels"]]
        graph_data = self.__gnn_model.tensorize(
            GraphData(
                node_information=graph_nodes,
                edges=datapoint["edges"],
                reference_nodes={"backbone_nodes": datapoint["backbone_sequence"]},
            )
        )
        if graph_data is None:
            return None  # Discard example

        target_data = self.__decoder_model.tensorize(
            DecoderData(
                input_elements=[graph_nodes[k] for k in datapoint["backbone_sequence"]],
                target_data=datapoint["method_name"],
            )
        )

        return TensorizedGraph2Seq(encoder_data=graph_data, decoder_data=target_data)

    def initialize_minibatch(self) -> Dict[str, Any]:
        return {
            "encoder_mb_data": self.__gnn_model.initialize_minibatch(),
            "decoder_mb_data": self.__decoder_model.initialize_minibatch(),
        }

    def extend_minibatch_with(
        self, tensorized_datapoint: TensorizedGraph2Seq, partial_minibatch: Dict[str, Any]
    ) -> bool:
        continue_adding = self.__gnn_model.extend_minibatch_with(
            tensorized_datapoint.encoder_data, partial_minibatch["encoder_mb_data"]
        )
        continue_adding &= self.__decoder_model.extend_minibatch_with(
            tensorized_datapoint.decoder_data, partial_minibatch["decoder_mb_data"]
        )
        return continue_adding

    def finalize_minibatch(
        self, accumulated_minibatch_data: Dict[str, Any], device: Union[str, torch.device]
    ) -> Dict[str, Any]:
        return {
            "encoder_mb_data": self.__gnn_model.finalize_minibatch(
                accumulated_minibatch_data["encoder_mb_data"], device=device
            ),
            "decoder_mb_data": self.__decoder_model.finalize_minibatch(
                accumulated_minibatch_data["decoder_mb_data"], device=device
            ),
        }

    def greedy_decode(
        self, data: List[CodeGraph2Seq], trained_network: Graph2SeqModule, device: Any
    ) -> List[Tuple[List[str], float]]:
        decoded_sequences = []
        for mb_data, input_data in self.minibatch_iterator(
            self.tensorize_dataset(iter(data), return_input_data=True),
            device,
            max_minibatch_size=50,
        ):
            input_concrete_values: List[str] = []
            for sample in input_data:
                sample = enforce_not_None(sample)
                input_concrete_values.extend(
                    sample["node_labels"][k].lower() for k in sample["backbone_sequence"]
                )

            with torch.no_grad():
                gnn_output = trained_network._gnn(**mb_data["encoder_mb_data"])  # type: GnnOutput
                mb_outputs = self.__decoder_model.greedy_decode(
                    input_concrete_values=input_concrete_values,
                    input_memories=gnn_output.output_node_representations[
                        gnn_output.node_idx_references["backbone_nodes"]
                    ],
                    input_memories_origin_idx=gnn_output.node_graph_idx_reference["backbone_nodes"],
                    initial_states=trained_network._get_initial_decoder_states(gnn_output),
                    neural_model=trained_network._decoder,
                )
                decoded_sequences.extend(mb_outputs)

        assert len(decoded_sequences) == len(data)
        return decoded_sequences



import numpy as np
from dpu_utils.utils import RichPath
from typing import Dict, List


class PPIGraphSample:
    """Data structure holding a single PPI graph."""

    def __init__(
        self,
        adjacency_lists: List[np.ndarray],
        node_features: np.ndarray,
        node_labels: np.ndarray,
    ):
        self._adjacency_lists = adjacency_lists
        self._node_features = node_features
        self._node_labels = node_labels

    @property
    def node_labels(self) -> np.ndarray:
        """Node labels to predict as ndarray of shape [V, C]"""
        return self._node_labels

    @property
    def adjacency_lists(self) -> List[np.ndarray]:
        """Adjacency information by edge type as list of ndarrays of shape [E, 2]"""
        return self._adjacency_lists

    @property
    def node_features(self) -> np.ndarray:
        """Initial node features as ndarray of shape [V, ...]"""
        return self._node_features


class PPIDatasetLoader:
    @classmethod
    def load_data(cls, data_dir: RichPath, data_fold: str) -> List[PPIGraphSample]:

        print(" Loading PPI %s data from %s." % (data_fold, data_dir))

        graph_json_data = data_dir.join("%s_graph.json" % data_fold).read_by_file_suffix()
        node_to_features = data_dir.join("%s_feats.npy" % data_fold).read_by_file_suffix()
        node_to_labels = data_dir.join("%s_labels.npy" % data_fold).read_by_file_suffix()
        node_to_graph_id = data_dir.join("%s_graph_id.npy" % data_fold).read_by_file_suffix()

        # We read in all the data in two steps:
        #  (1) Read features, labels. Implicitly, this gives us the number of nodes per graph.
        #  (2) Read all edges, and shift them so that each graph starts with node 0.
        fwd_edge_type = 0

        graph_id_to_graph_data: Dict[int, PPIGraphSample] = {}
        graph_id_to_node_offset: Dict[int, int] = {}
        num_total_nodes = node_to_features.shape[0]
        for node_id in range(num_total_nodes):
            graph_id = node_to_graph_id[node_id]
            # In case we are entering a new graph, note its ID, so that we can normalise everything to start at 0
            if graph_id not in graph_id_to_graph_data:
                graph_id_to_graph_data[graph_id] = PPIGraphSample(
                    adjacency_lists=[[]],
                    node_features=[],
                    node_labels=[],
                )
                graph_id_to_node_offset[graph_id] = node_id
            cur_graph_data = graph_id_to_graph_data[graph_id]
            cur_graph_data.node_features.append(node_to_features[node_id])
            cur_graph_data.node_labels.append(node_to_labels[node_id])

        for edge_info in graph_json_data["links"]:
            src_node, tgt_node = edge_info["source"], edge_info["target"]
            # First, shift node IDs so that each graph starts at node 0:
            graph_id = node_to_graph_id[src_node]
            graph_node_offset = graph_id_to_node_offset[graph_id]
            src_node, tgt_node = src_node - graph_node_offset, tgt_node - graph_node_offset

            cur_graph_data = graph_id_to_graph_data[graph_id]
            cur_graph_data.adjacency_lists[fwd_edge_type].append((src_node, tgt_node))

        final_graphs = []
        for graph_data in graph_id_to_graph_data.values():
            # numpy-ize:
            adj_lists = [np.array(graph_data.adjacency_lists[fwd_edge_type], dtype=np.int32)]
            final_graphs.append(
                PPIGraphSample(
                    adjacency_lists=adj_lists,
                    node_features=np.array(graph_data.node_features, dtype=np.float32),
                    node_labels=np.array(graph_data.node_labels, dtype=np.bool),
                )
            )

        return final_graphs
